"""Read and define addresses and paths for the hub."""

from __future__ import annotations
import os
import sys
import glob
import ipaddress
import json


# creating data directory within module
MODULE_DIR = os.path.dirname(__file__)
PACKAGE_DIR = os.path.dirname(MODULE_DIR)
DATA_DIR = os.path.join(PACKAGE_DIR, 'data')
HUB_DATA_DIR = os.path.join(DATA_DIR, 'hub')
BACKUP_DIR = os.path.join(HUB_DATA_DIR, 'backup')
IP_PATH = os.path.join(HUB_DATA_DIR, 'ip_addresses.json')
LOG_PATH = os.path.join(HUB_DATA_DIR, 'app.logs')
RECACHE_INTERVAL = 60 * 60  # update every hour
if not os.path.exists(DATA_DIR):
  os.mkdir(DATA_DIR)
if not os.path.exists(HUB_DATA_DIR):
  os.mkdir(HUB_DATA_DIR)
if not os.path.exists(BACKUP_DIR):
  os.mkdir(BACKUP_DIR)


def set_ips(path: str):
  """Verify and save device IP addresses in data directory."""
  with open(path) as f:
    addresses = f.read().splitlines()
  for a in addresses:
    try:
      ipaddress.ip_address(a)
    except ValueError as e:
      if a != 'self':
        raise e
  with open(IP_PATH, 'w') as f:
    json.dump(addresses, f)


def load_ips() -> list[str]:
  """Read IP addresses saved in data directory."""
  with open(IP_PATH) as f:
    return json.load(f)


if __name__ == '__main__':
  if len(sys.argv) < 2:
    raise ValueError('Expected path/to/ip_addresses.txt as additional arg!')
  arg = sys.argv[1]
  if arg == '--remove-addresses':
    os.remove(IP_PATH)
    print(f'Removed stored file at {IP_PATH}')
  elif arg == '--remove-data':
    for data_file in glob.glob(DATA_DIR + '*.csv'):
      os.remove(data_file)
    print('Removed all stored data.')
  elif arg == '--remove-logs':
    os.remove(LOG_PATH)
    print(f'Removed stored file at {LOG_PATH}')

  else:
    set_ips(arg)
    print('Successfully set IP addresses')



from bairy.device.validate import random_configs
from bairy.device.device import initialize_device


def test_device():
  """Load random configs and text device."""
  d = random_configs()
  _, sensors = initialize_device(d)
  for s in sensors:
    r1 = s.read()
    assert len(r1.values()) == 1
    r1 = list(r1.values())[0]
    assert r1 is not None

    r2 = s.read()
    assert len(r2.values()) == 1
    r2 = list(r2.values())[0]
    assert r2 is not None

    assert abs(r1 - r2) < 5



"""Create dash plot and dash table as endpoints."""

from __future__ import annotations
import os
import logging
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import plotly.io as pio
from dash import Dash
from dash_table import DataTable
import dash_core_components as dcc
import dash_html_components as html
from bairy.device import configs


pio.templates.default = 'plotly_white'


def determine_plot_configs():
  """Determine plotly axes and variables configurations."""

  d = configs.load_device()
  # the order of the dictionaries below matter for plot tracing
  # order is least important to most important
  sensor_headers: dict[str, list[str]] = {'random': [],
                                          'digital': [],
                                          'air': []}
  sensor_units = {'random': 'random',
                  'digital': 'intensity',
                  'air': 'micrograms / cubic meter'}

  for s in d.sensors:
    if s.sensor_type == 'air':
      sensor_headers['air'] += ['pm_1.0', 'pm_2.5', 'pm_10']
    else:
      sensor_headers[s.sensor_type].append(s.header)

  # cleanup dictionaries
  empties = [k for k in sensor_headers.keys() if sensor_headers[k] == []]
  for k in empties:
    del sensor_headers[k]
    del sensor_units[k]

  # keeping at most two types of headers
  if len(sensor_headers) == 3:
    del sensor_headers['random']
    del sensor_units['random']

  return sensor_headers, sensor_units


def preprocess_df(time_period: str = 'all', as_str: bool = False):
  """Preprocess pandas DataFrame."""
  df = pd.read_csv(configs.DATA_PATH)

  df = df.set_index('time')
  df.index = pd.to_datetime(df.index)

  sensor_headers, _ = determine_plot_configs()
  cols_to_keep = [col for key in sensor_headers for col in sensor_headers[key]]
  df = df[cols_to_keep]

  if time_period == 'day':
    start = pd.Timestamp.now() - pd.Timedelta('1 day')
    df = df[df.index > start]
  elif time_period == 'week':
    start = pd.Timestamp.now() - pd.Timedelta('7 days')
    df = df[df.index > start]

  df = resample_df(df)

  if as_str:  # for plotly table
    df = df.iloc[::-1]
    df.index = df.index.astype(str)
    df = df.round(3)
  return df.reset_index()  # move time back as a column


def resample_df(df):
  """Smooth and condense df by resampling."""
  # 60 has many divisors
  rules = ['1T', '2T', '3T', '4T', '5T', '6T', '10T', '20T', '30T', '1H']

  if len(df) > 300:  # only resample if df large enough
    for r in rules:
      df = df.resample(r).mean()
      if len(df) < 5000:
        break

  # smoothing even more
  df = df.rolling(7, center=True, min_periods=1).mean()
  return df


def create_fig(time_period: str = 'all'):
  """Create plotly figure using one or two y-axes."""

  # avoiding errors before any data is captured
  if not os.path.exists(configs.DATA_PATH):
    return go.Figure()

  try:
    fig = go.Figure()
    # see https://plotly.com/python/discrete-color/
    colors = iter(px.colors.qualitative.Bold)
    df = preprocess_df(time_period)
    sensor_headers, sensor_units = determine_plot_configs()

    for i, (key, cols) in enumerate(sensor_headers.items()):
      unit = sensor_units[key]
      if len(sensor_headers) == 2 and i == 0:
        side = 'right'
        opacity = 0.4
      else:
        side = 'left'
        opacity = 1.0

      yaxis = 'y' if i == 0 else 'y2'
      yaxis_layout = 'yaxis' if i == 0 else 'yaxis2'
      layout_params = {yaxis_layout: {'title': unit, 'side': side}}
      if i == 1:
        layout_params[yaxis_layout]['overlaying'] = 'y'

      for col in cols:
        fig.add_trace(go.Scatter(
            x=df['time'],
            y=df[col],
            name=col,
            opacity=opacity,
            line={'color': next(colors)},
            yaxis=yaxis))
      fig.update_layout(**layout_params)

    # threshold for pm2.5
    if 'air' in sensor_headers and df['pm_2.5'].max() > 40:
      yaxis = 'y2' if len(sensor_headers) == 2 else 'y'
      fig.add_trace(go.Scatter(
          x=df['time'],
          y=[35] * len(df),
          name='pm_2.5 safe threshold',
          yaxis=yaxis,
          line={'dash': 'dot', 'color': next(colors)}))

    name = configs.load_device().name
    if time_period == 'day':
      title = f'{name} data over last 24 hours'
    elif time_period == 'week':
      title = f'{name} data over last 7 days'
    else:
      title = f'{name} data over entire runtime'

    fig.update_layout(height=800, title=title)
    fig.layout.xaxis.rangeslider.visible = True
    fig.layout.yaxis.fixedrange = False
    return fig

  except (KeyError, FileNotFoundError) as e:
    logging.info(e)
    return go.Figure()


def serve_plot():
  """Dynamically serve dash_plot.layout."""

  fig_day = create_fig('day')
  fig_week = create_fig('week')
  fig_all = create_fig()

  return html.Div(children=[
      html.Div(children=[
          html.H1(children='bairy', style={'fontWeight': 'bold'}),
          html.H3(children='Display sensor data from Raspberry Pi.'),
          html.A(
              'about bairy',
              href='https://github.com/zebengberg/bairy',
              style={'margin': '20px'}),
          html.A(
              'table',
              href='/table',
              style={'margin': '20px'})]),
      dcc.Graph(id='graph_day', figure=fig_day),
      dcc.Graph(id='graph_week', figure=fig_week),
      dcc.Graph(id='graph_all', figure=fig_all)])


css = 'https://codepen.io/chriddyp/pen/bWLwgP.css'
dash_plot = Dash(requests_pathname_prefix='/plot/', external_stylesheets=[css])
# See https://dash.plotly.com/live-updates
dash_plot.layout = serve_plot


def serve_table():
  """Dynamically serve updated dash_table.layout."""

  # avoiding errors before any data is captured
  if not os.path.exists(configs.DATA_PATH):
    return DataTable()

  try:
    df = preprocess_df('day', True)
    columns = [{'name': i, 'id': i} for i in df.columns]
    data = df.to_dict('records')
    table = DataTable(
        id='table',
        columns=columns,
        data=data,
        style_cell=dict(textAlign='left'),
        style_header=dict(backgroundColor="paleturquoise"),
        style_data=dict(backgroundColor="lavender"))

  except (KeyError, FileNotFoundError) as e:
    logging.info(e)
    table = DataTable()

  return html.Div(children=[
      html.Div(children=[
          html.H1(children='bairy', style={'fontWeight': 'bold'}),
          html.H3(children='Display sensor data from Raspberry Pi.'),
          html.A(
              'about bairy',
              href='https://github.com/zebengberg/bairy',
              style={'margin': '20px'}),
          html.A(
              'plot',
              href='/plot',
              style={'margin': '20px'})]),
      table])


dash_table = Dash(requests_pathname_prefix='/table/',
                  external_stylesheets=[css])
# See https://dash.plotly.com/live-updates
dash_table.layout = serve_table



"""Process device requests with FastAPI."""


from __future__ import annotations
import os
import json
import subprocess
import sys
import logging
from fastapi import FastAPI
from fastapi import responses
from fastapi.middleware.wsgi import WSGIMiddleware
import uvicorn
from bairy import log_configs
from bairy.device import utils, configs, dash_app, device


app = FastAPI()
app.mount('/plot', WSGIMiddleware(dash_app.dash_plot.server))
app.mount('/table', WSGIMiddleware(dash_app.dash_table.server))


@app.get('/')
def root():
  """Redirect to plot."""
  return responses.RedirectResponse(url='/plot')


@app.get('/data')
def data():
  """Return streaming response of CSV containing all data."""
  # cannot use with ... here
  f = open(configs.DATA_PATH, 'rb')
  return responses.StreamingResponse(f, media_type='text/csv')


@app.get('/logs', response_class=responses.PlainTextResponse)
def logs():
  """Return app log as plain text."""
  with open(configs.LOG_PATH) as f:
    lines = f.readlines()
  # return as reverse chronological
  lines.reverse()
  return ''.join(lines)


@app.get('/status', response_class=responses.PlainTextResponse)
def status():
  """Return device status as plaintext json."""
  device_configs = configs.load_device().dict()
  size = utils.get_data_size()
  n_rows = utils.count_rows(configs.DATA_PATH)
  latest = utils.latest_data()
  ip_address = utils.get_local_ip_address()
  disk_space = utils.get_disk_space()
  bairy_version = utils.get_bairy_version()

  device_status = {'device_configs': device_configs,
                   'data_details': {'file_size': size, 'n_rows': n_rows},
                   'available_disk_space': disk_space,
                   'bairy_version': bairy_version,
                   'ip_address': ip_address,
                   'latest_reading': latest}
  return json.dumps(device_status, indent=4)


@app.get('/remote/{command}', response_class=responses.PlainTextResponse)
def remote(command: str):
  """Run remote command on device."""
  if command == 'reboot':
    logging.info('Reboot requested')
    try:
      subprocess.check_call(['sudo', 'reboot'])
    except subprocess.CalledProcessError as e:
      pass
    return 'rebooting'

  if command == 'update':
    logging.info('Update requested')
    call = [sys.executable, '-m', 'pip', 'install', '-U', 'bairy']
    try:
      subprocess.check_call(call)
      logging.info('Update success')
      return 'success'
    except subprocess.CalledProcessError as e:
      logging.error('Update failure')
      logging.error(e)
      return 'fail'

  elif command == 'remove-data':
    logging.info('Remove data requested')
    os.remove(configs.DATA_PATH)
    device.initialize_device()  # create new data.csv with headers
    return 'success'

  elif command == 'remove-logs':
    logging.info('Remove logs requested')
    with open(configs.LOG_PATH, 'w') as f:
      f.close()
    logging.info('Logs cleared')
    return 'success'

  return 'unknown command'


@app.post('/set-configs')
def set_configs(d: device.DeviceConfigs):
  """Set device config file remotely."""
  logging.info('setting new configs')
  logging.info(d)
  assert d == device.DeviceConfigs(**d.dict())
  with open(configs.CONFIGS_PATH, 'w') as f:
    json.dump(d.dict(), f, indent=4)
  return 'new configs will be active after reboot'


def run_app():
  """Run app with uvicorn."""
  uvicorn.run(
      app,
      host='0.0.0.0',
      port=8000,
      log_config=log_configs.get_uvicorn_logger(configs.LOG_PATH)
  )



"""Test all hub app endpoints."""

from fastapi.testclient import TestClient
from bairy.hub.app import app
from bairy.hub.request import validate_names
from bairy.hub.configs import load_ips


client = TestClient(app)


def test_ip_addresses_exists():
  """Confirm that the hub has stored ip addresses."""
  load_ips()
  validate_names()


def test_all_endpoint_status():
  """Test all endpoints status code through openapi.json json."""
  r = client.get('/openapi.json')
  assert r.status_code == 200
  for e in r.json()['paths'].keys():
    r = client.get(e)
    assert r.status_code == 200

  for e in ['plot']:
    r = client.get(e)
    assert r.status_code == 200



"""Test validate module."""

from bairy.device.validate import *


def test_sensor_configs():
  """Test various SensorConfigs models."""

  s = DigitalSensorConfigs(bcm_pin=17, header='random')
  assert s == DigitalSensorConfigs(**s.dict())

  s = RandomSensorConfigs(header='random')
  assert s == RandomSensorConfigs(**s.dict())

  s = AirSensorConfigs(i2c_address=0x12)
  assert s == AirSensorConfigs(**s.dict())


def test_device_configs():
  """Test DeviceConfigs model."""

  s1 = DigitalSensorConfigs(bcm_pin=17, header='random')
  s2 = RandomSensorConfigs(header='random')
  s3 = AirSensorConfigs(i2c_address=0x12)

  d = DeviceConfigs(
      name='razzy',
      location='garden',
      sensors=[s1, s2, s3],
      plot_axes={'foo': ['bar']},
      update_interval=17)

  assert d == DeviceConfigs(**d.dict())


def test_examples():
  """Use pydantic validator to test built in examples."""
  example_configs()
  random_configs()



"""Test __main__ functions."""

import os
from bairy.__main__ import parse_args, parse_device, parse_hub
from bairy import create_service


def test_parser():
  """Test argparse command line arguments."""
  args = parse_args(['--print-configs'])
  assert args.print_configs

  args = parse_args(['--print-data-path'])
  assert args.print_data_path

  args = parse_args(['--create-service'])
  assert args.create_service

  args = parse_args(['--set-random-configs'])
  assert args.set_random_configs

  args = parse_args(['--set-configs', 'fake_path'])
  assert args.path == ['fake_path']

  args = parse_args(['--remove', 'logs'])
  assert args.remove == ['logs']


def test_service():
  """Test create_service()."""
  create_service.create_service()


def test_main():
  """Simulate several main flags."""
  args = parse_args(['-p'])
  parse_device(args)
  parse_hub(args)

  args = parse_args(['-d'])
  parse_device(args)
  parse_hub(args)

  args = parse_args(['-t'])
  parse_device(args)
  os.remove('template_configs.json')

  try:
    args = parse_args(['-rm foo'])
    parse_device(args)
    parse_hub(args)
  except SystemExit:
    pass
  else:
    raise ValueError



"""Run app and read sensor data."""

from __future__ import annotations
import os
import shutil
import sys
import json
import glob
import argparse
import asyncio
from multiprocessing import Process
from bairy.device import configs, utils, app, device, validate
from bairy.hub import configs as hub_configs, app as hub_app, request
from bairy import create_service, log_configs


def parse_args(args: list[str]):
  """Create argparse parser."""
  parser = argparse.ArgumentParser()
  parser.add_argument(
      'mode',
      type=str,
      nargs='?',
      choices=['device', 'hub'],
      help='set the mode in which bairy runs',
      default='device')

  parser.add_argument(
      '-c',
      '--set-configs',
      type=str,
      nargs=1,
      dest='path',
      help='set configs (or IP addresses in hub mode) by passing path/to/configs.json',
      required=False)

  parser.add_argument(
      '-r',
      '--set-random-configs',
      action='store_true',
      help='set configs with random sensors',
      required=False)

  parser.add_argument(
      '-p',
      '--print-configs',
      action='store_true',
      help='print current configs',
      required=False)

  parser.add_argument(
      '-t',
      '--configs-template',
      action='store_true',
      help='create configs.json template',
      required=False)

  parser.add_argument(
      '-rm',
      '--remove',
      type=str,
      nargs=1,
      choices=['data', 'logs', 'configs', 'all', 'entire'],
      help='remove saved data',
      required=False)

  parser.add_argument(
      '-d',
      '--print-data-path',
      action='store_true',
      help='print path to data directory',
      required=False)

  parser.add_argument(
      '-s',
      '--create-service',
      action='store_true',
      help='create bairy.service file for systemd',
      required=False)

  return parser.parse_args(args)


def parse_device_remove(arg: str):
  """Parse --remove flag under device mode."""
  if arg in ['data', 'all']:
    if os.path.exists(configs.DATA_PATH):
      os.remove(configs.DATA_PATH)
      print(f'Removed stored file at {configs.DATA_PATH}')
  if arg in ['logs', 'all']:
    if os.path.exists(configs.LOG_PATH):
      os.remove(configs.LOG_PATH)
      print(f'Removed stored file at {configs.LOG_PATH}')
  if arg in ['configs', 'all']:
    if os.path.exists(configs.CONFIGS_PATH):
      os.remove(configs.CONFIGS_PATH)
      print(f'Removed stored file at {configs.CONFIGS_PATH}')
  if arg == 'entire':
    if os.path.exists(configs.DATA_DIR):
      shutil.rmtree(hub_configs.DATA_DIR)
      print(f'Removed entire directory {hub_configs.DATA_DIR}')


def parse_hub_remove(arg: str):
  """Parse --remove flag under hub mode."""
  if arg in ['data', 'all']:
    data_files = glob.glob(hub_configs.HUB_DATA_DIR + '/*.csv') + \
        glob.glob(hub_configs.BACKUP_DIR + '/*.csv')
    for f in data_files:
      os.remove(f)
      print(f'Removed stored file at {f}')
  if arg in ['logs', 'all']:
    if os.path.exists(hub_configs.LOG_PATH):
      os.remove(hub_configs.LOG_PATH)
      print(f'Removed stored file at {hub_configs.LOG_PATH}')
  if arg in ['configs', 'all']:
    if os.path.exists(hub_configs.IP_PATH):
      os.remove(hub_configs.IP_PATH)
      print(f'Removed stored file at {hub_configs.IP_PATH}')
  if arg == 'entire':
    shutil.rmtree(hub_configs.DATA_DIR)


def parse_device(args: argparse.Namespace):
  """Take actions under device mode."""
  if args.path:
    configs.set_configs(args.path[0])
  elif args.set_random_configs:
    configs.set_random_configs()
  elif args.print_configs:
    print(json.dumps(configs.load_device().dict(), indent=4))
  elif args.configs_template:
    with open('template_configs.json', 'w') as f:
      f.write(json.dumps(validate.example_configs().dict(), indent=4))
      print('Created template_configs.json file.')
  elif args.remove:
    parse_device_remove(args.remove[0])
  elif args.print_data_path:
    print(configs.DATA_PATH)
  elif args.create_service:
    create_service.create_service()
  else:
    if not os.path.exists(configs.CONFIGS_PATH):
      raise FileNotFoundError('No configurations found! Run bairy --help')

    print('#' * 65)
    print('LOCAL IP ADDRESS:', utils.get_local_ip_address())
    print('#' * 65)
    p = Process(target=app.run_app)
    p.start()
    asyncio.run(device.run_device())


def parse_hub(args: argparse.Namespace):
  """Take actions under hub mode."""
  if args.path:
    hub_configs.set_ips(args.path[0])
    request.validate_names()
  elif args.print_configs:
    print(json.dumps(hub_configs.load_ips(), indent=4))
  elif args.remove:
    parse_hub_remove(args.remove[0])
  elif args.create_service:
    create_service.create_service(True)
  elif args.print_data_path:
    print(hub_configs.HUB_DATA_DIR)
  elif args.configs_template or args.set_random_configs:
    raise NotImplementedError('Not available for hub mode.')
  else:
    if not os.path.exists(hub_configs.IP_PATH):
      raise FileNotFoundError('No IP addresses found! Run bairy --help')

    print('#' * 65)
    print('LOCAL IP ADDRESS:', utils.get_local_ip_address())
    print('#' * 65)
    p = Process(target=hub_app.run_app)
    p.start()

    ip_addresses = hub_configs.load_ips()
    if 'self' in ip_addresses:
      task = asyncio.gather(device.run_device(), request.run_requests())
      asyncio.run(task)
    else:
      asyncio.run(request.run_requests())


def main():
  """Parse command line arguments and run actions."""
  args = parse_args(sys.argv[1:])
  if args.mode == 'device':
    log_configs.configure_root_logging(configs.LOG_PATH)
    parse_device(args)
  else:
    assert args.mode == 'hub'
    log_configs.configure_root_logging(hub_configs.LOG_PATH)
    parse_hub(args)


if __name__ == '__main__':
  main()



"""Dash app to plot device data."""

import os
import glob
import logging
import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
from dash import Dash
import dash_core_components as dcc
import dash_html_components as html
from bairy.hub.configs import BACKUP_DIR, HUB_DATA_DIR, load_ips, IP_PATH
from bairy.device import configs as device_configs
from bairy.device.dash_app import resample_df


def load_data(only_last_day: bool):
  """Load cached data."""

  ip_addresses = load_ips()
  dfs = glob.glob(HUB_DATA_DIR + '/*.csv')
  names = [f.split('/')[-1].split('.')[0] for f in dfs]

  # looking for copy in hub/backup to use instead
  for i, name in enumerate(names):
    backup_csv = os.path.join(BACKUP_DIR, name + '.csv')
    if os.path.exists(backup_csv):
      dfs[i] = backup_csv

  if 'self' in ip_addresses and os.path.exists(device_configs.DATA_PATH):
    dfs.append(device_configs.DATA_PATH)
    names.append(device_configs.load_device().name)

  dfs = [pd.read_csv(f) for f in dfs]

  # must include time
  cols_to_keep = ['time', 'pm_2.5', 'pm_1.0']
  dfs = [df[cols_to_keep]
         for df in dfs if set(cols_to_keep).issubset(set(df.columns))]

  dfs = [df.set_index(pd.to_datetime(df['time'])) for df in dfs]

  # dfs = [df.drop('time', axis=1) for df in dfs]
  dfs = [df.rename(columns={k: name + ' ' + k for k in df.columns})
         for df, name in zip(dfs, names)]

  df = pd.concat(dfs)
  if only_last_day:
    start = pd.Timestamp.now() - pd.Timedelta('1 day')
    df = df[df.index > start]
  df = resample_df(df)
  return df


def create_fig(only_last_day: bool):
  """Create plotly figure using one or two y-axes."""

  # avoiding errors when device not configured to run as hub
  if not os.path.exists(IP_PATH):
    return px.line()

  try:
    df = load_data(only_last_day)
    fig = px.line(df, x=df.index, y=df.columns)

    # showing pm2.5 safe threshold in any column exceeds
    for k in df.columns:
      if 'pm_2.5' in k:
        if df[k].max() > 40:
          fig.add_trace(go.Scatter(
              x=df.index,
              y=[35] * len(df),
              name='pm_2.5 safe threshold',
              line={'dash': 'dot'}))
          break

    if only_last_day:
      title = 'Data over last 24 hours'
    else:
      title = 'Data over entire runtime'
    fig.update_layout(height=800, title=title)
    fig.layout.xaxis.rangeslider.visible = True
    fig.layout.yaxis.title = 'micrograms / cubic meter'
    return fig

  except (KeyError, FileNotFoundError, ValueError) as e:
    logging.info(e)
    return px.line()


def serve_plot():
  """Serve dash app plot."""
  fig_day = create_fig(True)
  fig_all = create_fig(False)

  return html.Div(children=[
      html.Div(children=[
          html.H1(children='bairy', style={'fontWeight': 'bold'}),
          html.H3(children='Display sensor data from Raspberry Pi.'),
          html.A(
              'about bairy',
              href='https://github.com/zebengberg/bairy',
              style={'margin': '20px'})]),
      dcc.Graph(id='graph_day', figure=fig_day),
      dcc.Graph(id='graph_all', figure=fig_all)])


css = 'https://codepen.io/chriddyp/pen/bWLwgP.css'
dash_plot = Dash(requests_pathname_prefix='/plot/', external_stylesheets=[css])
dash_plot.layout = serve_plot



"""Configure logging."""

from __future__ import annotations
from typing import Any
import logging
import uvicorn


DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
LOG_FORMAT = '%(asctime)s - %(levelname)s - %(message)s'


def configure_root_logging(log_path: str):
  """Set up root logger to send logs to file and console."""
  logger = logging.getLogger()
  logger.setLevel(logging.INFO)
  formatter = logging.Formatter(
      fmt=LOG_FORMAT,
      datefmt=DATE_FORMAT
  )
  file_handler = logging.FileHandler(log_path)
  file_handler.setFormatter(formatter)
  logger.addHandler(file_handler)
  console_handler = logging.StreamHandler()
  console_handler.setFormatter(formatter)
  logger.addHandler(console_handler)


def get_uvicorn_logger(log_path: str):
  """Configure uvicorn default logger to save to file."""

  lc: dict[str, Any] = uvicorn.config.LOGGING_CONFIG

  # changing default formatting
  lc['formatters']['access']['fmt'] = LOG_FORMAT
  lc['formatters']['access']['datefmt'] = DATE_FORMAT
  lc['formatters']['default']['fmt'] = LOG_FORMAT
  lc['formatters']['default']['datefmt'] = DATE_FORMAT

  # adding FileHandler to LOGGING_CONFIG
  lc['handlers']['default_file'] = {
      'formatter': 'default',
      'class': 'logging.FileHandler',
      'filename': log_path}
  lc['handlers']['access_file'] = {
      'formatter': 'access',
      'class': 'logging.FileHandler',
      'filename': log_path}

  # telling loggers to use the additional handler
  lc['loggers']['uvicorn']['handlers'].append('default_file')
  lc['loggers']['uvicorn.access']['handlers'].append('access_file')
  return lc



"""FastAPI app to display device data."""

import json
from fastapi import FastAPI
from fastapi.responses import PlainTextResponse, RedirectResponse
from fastapi.middleware.wsgi import WSGIMiddleware
import uvicorn
from bairy.hub import configs
from bairy.hub.request import get_all_statuses
from bairy.hub.dash_plot import dash_plot
from bairy import log_configs


app = FastAPI()
app.mount('/plot', WSGIMiddleware(dash_plot.server))


@app.get('/')
def root():
  """Redirect to plot."""
  return RedirectResponse(url='/plot')


@app.get('/status', response_class=PlainTextResponse)
def status():
  """Get status of each device."""
  statuses = get_all_statuses()
  return json.dumps(statuses, indent=4)


@app.get('/logs', response_class=PlainTextResponse)
def logs():
  """Return app log as plain text."""
  with open(configs.LOG_PATH) as f:
    return f.read()


def run_app():
  """Run app as separate process."""
  uvicorn.run(
      app,
      host='0.0.0.0',
      port=8000,
      log_config=log_configs.get_uvicorn_logger(configs.LOG_PATH)
  )
